@@ -19,16 +19,17 @@
  */
 
 #include <config.h>
-#include "exif-mnote-data.h"
 
-#include "exif-data.h"
-#include "exif-ifd.h"
-#include "exif-mnote-data-priv.h"
-#include "exif-utils.h"
-#include "exif-loader.h"
-#include "jpeg-marker.h"
+#include <libexif/exif-mnote-data.h>
+#include <libexif/exif-data.h>
+#include <libexif/exif-ifd.h>
+#include <libexif/exif-mnote-data-priv.h>
+#include <libexif/exif-utils.h>
+#include <libexif/exif-loader.h>
 #include <libexif/exif-log.h>
 
+#include <libjpeg/jpeg-marker.h>
+
 #include <libexif/olympus/exif-mnote-data-olympus.h>
 #include <libexif/canon/exif-mnote-data-canon.h>
 #include <libexif/pentax/exif-mnote-data-pentax.h>
@@ -40,7 +41,9 @@
 #undef MAX
 #define MAX(a, b)  (((a) > (b)) ? (a) : (b))
 
-/* #define DEBUG */
+#if defined(__WATCOMC__) || defined(_MSC_VER)
+#      define strncasecmp strnicmp
+#endif
 
 static const unsigned char ExifHeader[] = {0x45, 0x78, 0x69, 0x66, 0x00, 0x00};
 
@@ -51,6 +54,7 @@ struct _ExifDataPrivate
 	ExifMnoteData *md;
 
 	ExifLog *log;
+	ExifMem *mem;
 
 	unsigned int ref_count;
 
@@ -58,6 +62,20 @@ struct _ExifDataPrivate
 	unsigned int offset_mnote;
 };
 
+static void *
+exif_data_alloc (ExifData *data, unsigned int i)
+{
+	void *d;
+
+	if (!data || !i) return NULL;
+
+	d = exif_mem_alloc (data->priv->mem, i);
+	if (d) return d;
+
+	EXIF_LOG_NO_MEMORY (data->priv->log, "ExifData", i);
+	return NULL;
+}
+
 ExifMnoteData *
 exif_data_get_mnote_data (ExifData *d)
 {
@@ -66,24 +84,34 @@ exif_data_get_mnote_data (ExifData *d)
 
 ExifData *
 exif_data_new (void)
+{
+	ExifMem *mem = exif_mem_new_default ();
+	ExifData *d = exif_data_new_mem (mem);
+
+	exif_mem_unref (mem);
+
+	return d;
+}
+
+ExifData *
+exif_data_new_mem (ExifMem *mem)
 {
 	ExifData *data;
 	unsigned int i;
 
-	data = malloc (sizeof (ExifData));
-	if (!data)
-		return (NULL);
-	memset (data, 0, sizeof (ExifData));
-	data->priv = malloc (sizeof (ExifDataPrivate));
-	if (!data->priv) {
-		free (data);
-		return (NULL);
-	}
-	memset (data->priv, 0, sizeof (ExifDataPrivate));
+	if (!mem) return NULL;
+
+	data = exif_mem_alloc (mem, sizeof (ExifData));
+	if (!data) return (NULL);
+	data->priv = exif_mem_alloc (mem, sizeof (ExifDataPrivate));
+	if (!data->priv) { exif_mem_free (mem, data); return (NULL); }
 	data->priv->ref_count = 1;
 
+	data->priv->mem = mem;
+	exif_mem_ref (mem);
+
 	for (i = 0; i < EXIF_IFD_COUNT; i++) {
-		data->ifd[i] = exif_content_new ();
+		data->ifd[i] = exif_content_new_mem (data->priv->mem);
 		if (!data->ifd[i]) {
 			exif_data_free (data);
 			return (NULL);
@@ -135,17 +163,16 @@ exif_data_load_data_entry (ExifData *data, ExifEntry *entry,
 	if (size < doff + s)
 		return;
 
-	entry->data = malloc (s);
-	if (!entry->data) {
-		EXIF_LOG_NO_MEMORY (data->priv->log, "ExifData", s);
-		return;
+	entry->data = exif_data_alloc (data, s);
+	if (entry->data) {
+		entry->size = s;
+		memcpy (entry->data, d + doff, s);
 	}
-	entry->size = s;
-	memcpy (entry->data, d + doff, s);
 
 	/* If this is the MakerNote, remember the offset */
 	if (entry->tag == EXIF_TAG_MAKER_NOTE) {
-		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		if (entry->size > 6) exif_log (data->priv->log,
+			  EXIF_LOG_CODE_DEBUG, "ExifData",
 		          "MakerNote found (%02x %02x %02x %02x "
 			  "%02x %02x %02x...).",
 			  entry->data[0], entry->data[1], entry->data[2],
@@ -153,6 +180,8 @@ exif_data_load_data_entry (ExifData *data, ExifEntry *entry,
 			  entry->data[6]);
 		data->priv->offset_mnote = doff;
 	}
+
+	exif_entry_fix (entry);
 }
 
 static void
@@ -162,6 +191,8 @@ exif_data_save_data_entry (ExifData *data, ExifEntry *e,
 {
 	unsigned int doff, s;
 
+	if (!data || !data->priv) return;
+
 	/*
 	 * Each entry is 12 bytes long. The memory for the entry has
 	 * already been allocated.
@@ -171,15 +202,17 @@ exif_data_save_data_entry (ExifData *data, ExifEntry *e,
 	exif_set_short (*d + 6 + offset + 2,
 			data->priv->order, (ExifShort) e->format);
 
+#ifndef EXIF_DONT_CHANGE_MAKER_NOTE
 	/* If this is the maker note tag, update it. */
 	if ((e->tag == EXIF_TAG_MAKER_NOTE) && data->priv->md) {
-		free (e->data);
+		exif_mem_free (data->priv->mem, e->data);
 		e->data = NULL;
 		e->size = 0;
 		exif_mnote_data_set_offset (data->priv->md, *ds - 6);
 		exif_mnote_data_save (data->priv->md, &e->data, &e->size);
 		e->components = e->size;
 	}
+#endif
 
 	exif_set_long  (*d + 6 + offset + 4,
 			data->priv->order, e->components);
@@ -192,13 +225,21 @@ exif_data_save_data_entry (ExifData *data, ExifEntry *e,
 	if (s > 4) {
 		doff = *ds - 6;
 		*ds += s;
-		*d = realloc (*d, *ds);
+
+		/*
+		 * According to the TIFF specification,
+		 * the offset must be an even number. If we need to introduce
+		 * a padding byte, we set it to 0.
+		 */
+		if (s & 1) *ds++;
+		*d = exif_mem_realloc (data->priv->mem, *d, *ds);
 		if (!*d) {
 			EXIF_LOG_NO_MEMORY (data->priv->log, "ExifData", *ds);
 		  	return;
 		}
-		exif_set_long (*d + 6 + offset + 8,
-			       data->priv->order, doff);
+		exif_set_long (*d + 6 + offset + 8, data->priv->order, doff);
+		if (s & 1) *(*d + *ds - 1) = '\0';
+
 	} else
 		doff = offset + 8;
 
@@ -217,17 +258,31 @@ exif_data_load_data_thumbnail (ExifData *data, const unsigned char *d,
 			  (int) ds, (int) offset, (int) size);
 		return;
 	}
-	if (data->data)
-		free (data->data);
+	if (data->data) exif_mem_free (data->priv->mem, data->data);
 	data->size = size;
-	data->data = malloc (data->size);
-	if (!data->data) {
-		EXIF_LOG_NO_MEMORY (data->priv->log, "ExifData", data->size);
-		return;
-	}
+	data->data = exif_data_alloc (data, data->size);
+	if (!data->data) return;
 	memcpy (data->data, d + offset, data->size);
 }
 
+#undef CHECK_REC
+#define CHECK_REC(i) 					\
+if (data->ifd[(i)] == ifd) {				\
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, \
+		"ExifData", "Recursive entry in IFD "	\
+		"'%s' detected. Skipping...",		\
+		exif_ifd_get_name (i));			\
+	break;						\
+}							\
+if (data->ifd[(i)]->count) {				\
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG,	\
+		"ExifData", "Attemt to load IFD "	\
+		"'%s' multiple times detected. "	\
+		"Skipping...",				\
+		exif_ifd_get_name (i));			\
+	break;						\
+}
+
 static void
 exif_data_load_data_content (ExifData *data, ExifContent *ifd,
 			     const unsigned char *d,
@@ -239,6 +294,8 @@ exif_data_load_data_content (ExifData *data, ExifContent *ifd,
 	unsigned int i;
 	ExifTag tag;
 
+	if (!data || !data->priv) return;
+
 	/* Read the number of entries */
 	if (offset >= ds - 1) return;
 	n = exif_get_short (d + offset, data->priv->order);
@@ -262,21 +319,21 @@ exif_data_load_data_content (ExifData *data, ExifContent *ifd,
 					   data->priv->order);
 			switch (tag) {
 			case EXIF_TAG_EXIF_IFD_POINTER:
+				CHECK_REC (EXIF_IFD_EXIF);
 				exif_data_load_data_content (data,
 					data->ifd[EXIF_IFD_EXIF], d, ds, o);
 				break;
 			case EXIF_TAG_GPS_INFO_IFD_POINTER:
+				CHECK_REC (EXIF_IFD_GPS);
 				exif_data_load_data_content (data,
 					data->ifd[EXIF_IFD_GPS], d, ds, o);
 				break;
 			case EXIF_TAG_INTEROPERABILITY_IFD_POINTER:
+				CHECK_REC (EXIF_IFD_INTEROPERABILITY);
 				exif_data_load_data_content (data,
 					data->ifd[EXIF_IFD_INTEROPERABILITY], d, ds, o);
 				break;
 			case EXIF_TAG_JPEG_INTERCHANGE_FORMAT:
-#ifdef DEBUG
-				printf ("Thumbnail at %i.\n", (int) o);
-#endif
 				thumbnail_offset = o;
 				if (thumbnail_offset && thumbnail_length)
 					exif_data_load_data_thumbnail (data, d,
@@ -284,9 +341,6 @@ exif_data_load_data_content (ExifData *data, ExifContent *ifd,
 						thumbnail_length);
 				break;
 			case EXIF_TAG_JPEG_INTERCHANGE_FORMAT_LENGTH:
-#ifdef DEBUG
-				printf ("Thumbnail size: %i.\n", (int) o);
-#endif
 				thumbnail_length = o;
 				if (thumbnail_offset && thumbnail_length)
 					exif_data_load_data_thumbnail (data, d,
@@ -300,14 +354,18 @@ exif_data_load_data_content (ExifData *data, ExifContent *ifd,
 		default:
 
 			/*
-			 * If we don't know the tag, changes are high
+			 * If we don't know the tag, chances are high
 			 * that the EXIF data does not follow the standard.
 			 */
-			if (!exif_tag_get_name (tag)) return;
-			entry = exif_entry_new ();
-			exif_content_add_entry (ifd, entry);
+			if (!exif_tag_get_name (tag)) {
+				exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+				  "Unknown tag %x (entry %i)", tag, i);
+				return;
+			}
+			entry = exif_entry_new_mem (data->priv->mem);
 			exif_data_load_data_entry (data, entry, d, ds,
 						   offset + 12 * i);
+			exif_content_add_entry (ifd, entry);
 			exif_entry_unref (entry);
 			break;
 		}
@@ -321,9 +379,9 @@ exif_data_save_data_content (ExifData *data, ExifContent *ifd,
 {
 	unsigned int j, n_ptr = 0, n_thumb = 0;
 	ExifIfd i;
+	ExifEntry *le = NULL;
 
-	if (!data || !ifd || !d || !ds)
-		return;
+	if (!data || !data->priv || !ifd || !d || !ds) return;
 
 	for (i = 0; i < EXIF_IFD_COUNT; i++)
 		if (ifd == data->ifd[i])
@@ -366,7 +424,7 @@ exif_data_save_data_content (ExifData *data, ExifContent *ifd,
 	 * and the number of entries.
 	 */
 	*ds += (2 + (ifd->count + n_ptr + n_thumb) * 12 + 4);
-	*d = realloc (*d, *ds);
+	*d = exif_mem_realloc (data->priv->mem, *d, *ds);
 	if (!*d) {
 		EXIF_LOG_NO_MEMORY (data->priv->log, "ExifData", *ds);
 	  	return;
@@ -377,15 +435,32 @@ exif_data_save_data_content (ExifData *data, ExifContent *ifd,
 			(ExifShort) (ifd->count + n_ptr + n_thumb));
 	offset += 2;
 
-#ifdef DEBUG
-	printf ("Saving %i entries (IFD '%s', offset: %i)...\n",
-		ifd->count, exif_ifd_get_name (i), offset);
-#endif
+	/*
+	 * Save each entry. TIFF specification requires entries to be sorted
+	 * by tags. Yes, the sorting routine can be improved. And yes, the addition
+	 * of special entries at the end is wrong (they should be inserted at
+	 * the right position). Please submit a patch.
+	 */
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		  "Saving %i entries (IFD '%s', offset: %i)...",
+		  ifd->count, exif_ifd_get_name (i), offset);
+	for (j = 0; j < ifd->count; j++) {
+		ExifEntry *me = NULL;
+		unsigned int k;
+
+		for (k = 0; k < ifd->count; k++) {
+			if (le && ((ifd->entries[k]->tag < le->tag) ||
+						((ifd->entries[k]->tag == le->tag) && (le <= ifd->entries[k]))))
+				continue;
+			if (!me || (me && ((me->tag > ifd->entries[k]->tag) ||
+							((me->tag == ifd->entries[k]->tag) && (me <= ifd->entries[k])))))
+				me = ifd->entries[k];
+		}
+		if (!me) break;
+		exif_data_save_data_entry (data, me, d, ds, offset + 12 * j);
+		le = me;
+	}
 
-	/* Save each entry */
-	for (j = 0; j < ifd->count; j++)
-		exif_data_save_data_entry (data, ifd->entries[j],
-				d, ds, offset + 12 * j);
 	offset += 12 * ifd->count;
 
 	/* Now save special entries. */
@@ -469,7 +544,7 @@ exif_data_save_data_content (ExifData *data, ExifContent *ifd,
 			exif_set_long  (*d + 6 + offset + 8, data->priv->order,
 					*ds - 6);
 			*ds += data->size;
-			*d = realloc (*d, *ds);
+			*d = exif_mem_realloc (data->priv->mem, *d, *ds);
 			if (!*d) {
 				EXIF_LOG_NO_MEMORY (data->priv->log, "ExifData",
 						    *ds);
@@ -477,11 +552,6 @@ exif_data_save_data_content (ExifData *data, ExifContent *ifd,
 			}
 			memcpy (*d + *ds - data->size, data->data, data->size);
 			offset += 12;
-#ifdef DEBUG
-			printf ("Wrote %i bytes of thumbnail data at offset "
-				"%i.\n", data->size, *ds - data->size);
-			printf ("We currently have %i bytes EXIF data.\n", *ds);
-#endif
 
 			/* EXIF_TAG_JPEG_INTERCHANGE_FORMAT_LENGTH */
 			exif_set_short (*d + 6 + offset + 0, data->priv->order,
@@ -515,6 +585,49 @@ exif_data_save_data_content (ExifData *data, ExifContent *ifd,
 		exif_set_long (*d + 6 + offset, data->priv->order, 0);
 }
 
+typedef enum {
+	EXIF_DATA_TYPE_MAKER_NOTE_NONE		= 0,
+	EXIF_DATA_TYPE_MAKER_NOTE_CANON		= 1,
+	EXIF_DATA_TYPE_MAKER_NOTE_OLYMPUS	= 2,
+	EXIF_DATA_TYPE_MAKER_NOTE_PENTAX	= 3
+} ExifDataTypeMakerNote;
+
+static ExifDataTypeMakerNote
+exif_data_get_type_maker_note (ExifData *d)
+{
+	ExifEntry *e, *em;
+	char value[1024];
+
+	if (!d) return EXIF_DATA_TYPE_MAKER_NOTE_NONE;
+	
+	e = exif_data_get_entry (d, EXIF_TAG_MAKER_NOTE);
+        if (!e) return EXIF_DATA_TYPE_MAKER_NOTE_NONE;
+
+	/* Olympus & Nikon */
+	if ((e->size >= 5) && (!memcmp (e->data, "OLYMP", 5) ||
+			       !memcmp (e->data, "Nikon", 5)))
+		return EXIF_DATA_TYPE_MAKER_NOTE_OLYMPUS;
+
+	em = exif_data_get_entry (d, EXIF_TAG_MAKE);
+	if (!em) return EXIF_DATA_TYPE_MAKER_NOTE_NONE;
+
+	/* Canon */
+	if (!strcmp (exif_entry_get_value (em, value, sizeof(value)), "Canon"))
+		return EXIF_DATA_TYPE_MAKER_NOTE_CANON;
+
+	/* Pentax & some variant of Nikon */
+	if ((e->size >= 2) && (e->data[0] == 0x00) && (e->data[1] == 0x1b)) {
+		if (!strncasecmp (
+			exif_entry_get_value (em, value, sizeof(value)),
+					      "Nikon", 5))
+			return EXIF_DATA_TYPE_MAKER_NOTE_OLYMPUS;
+		else
+			return EXIF_DATA_TYPE_MAKER_NOTE_PENTAX;
+	}
+
+	return EXIF_DATA_TYPE_MAKER_NOTE_NONE;
+}
+
 void
 exif_data_load_data (ExifData *data, const unsigned char *d_orig,
 		     unsigned int ds_orig)
@@ -522,38 +635,27 @@ exif_data_load_data (ExifData *data, const unsigned char *d_orig,
 	unsigned int l;
 	ExifLong offset;
 	ExifShort n;
-	ExifEntry *e, *em;
 	const unsigned char *d = d_orig;
 	unsigned int ds = ds_orig, len;
 
-	if (!data)
-		return;
-	if (!d || !ds)
-		return;
+	if (!data || !data->priv || !d || !ds) return;
 
-#ifdef DEBUG
-	printf ("Parsing %i byte(s) EXIF data...\n", ds);
-#endif
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		  "Parsing %i byte(s) EXIF data...\n", ds);
 
 	/*
 	 * It can be that the data starts with the EXIF header. If it does
 	 * not, search the EXIF marker.
 	 */
 	if (ds < 6) {
-#ifdef DEBUG
-		printf ("Size too small.\n");
-#endif
+		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		          "Size too small.");
 		return;
 	}
 	if (!memcmp (d, ExifHeader, 6)) {
-#ifdef DEBUG
-		printf ("Found EXIF header.\n");
-#endif
+		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+			  "Found EXIF header.");
 	} else {
-#ifdef DEBUG
-		printf ("Data begins with 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x "
-			"0x%x...\n", d[0], d[1], d[2], d[3], d[4], d[5], d[6]);
-#endif
 		while (1) {
 			while ((d[0] == 0xff) && ds) {
 				d++;
@@ -584,23 +686,21 @@ exif_data_load_data (ExifData *data, const unsigned char *d_orig,
 				break;
 
 			/* Unknown marker or data. Give up. */
-#ifdef DEBUG
-			printf ("EXIF marker not found.\n");
-#endif
+			exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG,
+				  "ExifData", "EXIF marker not found.");
 			return;
 		}
 		d++;
 		ds--;
 		if (ds < 2) {
-#ifdef DEBUG
-			printf ("Size too small.\n");
-#endif
+			exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG,
+				  "ExifData", "Size too small.");
 			return;
 		}
 		len = (d[0] << 8) | d[1];
-#ifdef DEBUG
-		printf ("We have to deal with %i byte(s) of EXIF data.\n", len);
-#endif
+		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+			  "We have to deal with %i byte(s) of EXIF data.",
+			  len);
 		d += 2;
 		ds -= 2;
 	}
@@ -610,24 +710,21 @@ exif_data_load_data (ExifData *data, const unsigned char *d_orig,
 	 * (offset 2, length 6).
 	 */
 	if (ds < 6) {
-#ifdef DEBUG
-		printf ("Size too small.\n");
-#endif
+		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+			  "Size too small.");
 		return;
 	}
 	if (memcmp (d, ExifHeader, 6)) {
-#ifdef DEBUG
-		printf ("EXIF header not found.\n");
-#endif
+		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+			  "EXIF header not found.");
 		return;
 	}
 
-#ifdef DEBUG
-	printf ("Found EXIF header.\n");
-#endif
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		  "Found EXIF header.");
 
 	/* Byte order (offset 6, length 2) */
-	if (ds < 12)
+	if (ds < 14)
 		return;
 	if (!memcmp (d + 6, "II", 2))
 		data->priv->order = EXIF_BYTE_ORDER_INTEL;
@@ -642,27 +739,30 @@ exif_data_load_data (ExifData *data, const unsigned char *d_orig,
 
 	/* IFD 0 offset */
 	offset = exif_get_long (d + 10, data->priv->order);
-#ifdef DEBUG
-	printf ("IFD 0 at %i.\n", (int) offset);
-#endif
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData", 
+		  "IFD 0 at %i.", (int) offset);
 
-	/* Parse the actual exif data (offset 14) */
+	/* Parse the actual exif data (usually offset 14 from start) */
 	exif_data_load_data_content (data, data->ifd[EXIF_IFD_0], d + 6,
 				     ds - 6, offset);
 
 	/* IFD 1 offset */
+	if (offset + 6 + 2 > ds) {
+		return;
+	}
 	n = exif_get_short (d + 6 + offset, data->priv->order);
+	if (offset + 6 + 2 + 12 * n + 4 > ds) {
+		return;
+	}
 	offset = exif_get_long (d + 6 + offset + 2 + 12 * n, data->priv->order);
 	if (offset) {
-#ifdef DEBUG
-		printf ("IFD 1 at %i.\n", (int) offset);
-#endif
+		exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+			  "IFD 1 at %i.", (int) offset);
 
 		/* Sanity check. */
 		if (offset > ds - 6) {
-#ifdef DEBUG
-			printf ("Bogus offset!\n");
-#endif
+			exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, 
+				  "ExifData", "Bogus offset!");
 			return;
 		}
 
@@ -676,46 +776,30 @@ exif_data_load_data (ExifData *data, const unsigned char *d_orig,
 	 * space between IFDs. Here is the only place where we have access
 	 * to that data.
 	 */
-	e = exif_data_get_entry (data, EXIF_TAG_MAKER_NOTE);
-	if (e) {
+	switch (exif_data_get_type_maker_note (data)) {
+	case EXIF_DATA_TYPE_MAKER_NOTE_OLYMPUS:
+		data->priv->md = exif_mnote_data_olympus_new (data->priv->mem);
+		break;
+	case EXIF_DATA_TYPE_MAKER_NOTE_PENTAX:
+		data->priv->md = exif_mnote_data_pentax_new (data->priv->mem);
+		break;
+	case EXIF_DATA_TYPE_MAKER_NOTE_CANON:
+		data->priv->md = exif_mnote_data_canon_new (data->priv->mem);
+		break;
+	default:
+		break;
+	}
 
-	    /* Olympus & Nikon */
-	    if ((e->size >= 5) && (!memcmp (e->data, "OLYMP", 5) ||
-				   !memcmp (e->data, "Nikon", 5))) {
-			data->priv->md = exif_mnote_data_olympus_new ();
-	    } else {
-			char value[7];
-			em = exif_data_get_entry (data, EXIF_TAG_MAKE);
-			/* Pentax & some variant of Nikon */
-			if ((e->size >= 2) && (e->data[0] == 0x00)
-				    && (e->data[1] == 0x1b)) {
-				if (em && !strncasecmp (exif_entry_get_value (em, value, sizeof(value)), "Nikon", 5)) {
-					data->priv->md = exif_mnote_data_olympus_new ();
-				} else {
-					data->priv->md = exif_mnote_data_pentax_new ();
-				}
-
-			} else {
-				/* Canon */
-				if (em) {
-		
-					if (!strcmp (exif_entry_get_value (em, value, sizeof(value)), "Canon"))
-						data->priv->md = exif_mnote_data_canon_new ();
-				}
-			}
-		}
-
-	    /* 
-	     * If we are able to interpret the maker note, do so.
-	     */
-	    if (data->priv->md) {
-		    exif_mnote_data_log (data->priv->md, data->priv->log);
-		    exif_mnote_data_set_byte_order (data->priv->md,
-				    		    data->priv->order);
-		    exif_mnote_data_set_offset (data->priv->md,
-				    		data->priv->offset_mnote);
-		    exif_mnote_data_load (data->priv->md, d, ds);
-	    }
+	/* 
+	 * If we are able to interpret the maker note, do so.
+	 */
+	if (data->priv->md) {
+		exif_mnote_data_log (data->priv->md, data->priv->log);
+		exif_mnote_data_set_byte_order (data->priv->md,
+						data->priv->order);
+		exif_mnote_data_set_offset (data->priv->md,
+					    data->priv->offset_mnote);
+		exif_mnote_data_load (data->priv->md, d, ds);
 	}
 }
 
@@ -727,9 +811,8 @@ exif_data_save_data (ExifData *data, unsigned char **d, unsigned int *ds)
 
 	/* Header */
 	*ds = 14;
-	*d = malloc (*ds);
-	if (!*d)
-	  	return;
+	*d = exif_data_alloc (data, *ds);
+	if (!*d) return;
 	memcpy (*d, ExifHeader, 6);
 
 	/* Order (offset 6) */
@@ -751,15 +834,12 @@ exif_data_save_data (ExifData *data, unsigned char **d, unsigned int *ds)
 	exif_set_long (*d + 10, data->priv->order, 8);
 
 	/* Now save IFD 0. IFD 1 will be saved automatically. */
-#ifdef DEBUG
-	printf ("Saving IFDs...\n");
-#endif
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		  "Saving IFDs...");
 	exif_data_save_data_content (data, data->ifd[EXIF_IFD_0], d, ds,
 				     *ds - 6);
-
-#ifdef DEBUG
-	printf ("Saved %i byte(s) EXIF data.\n", *ds);
-#endif
+	exif_log (data->priv->log, EXIF_LOG_CODE_DEBUG, "ExifData",
+		  "Saved %i byte(s) EXIF data.", *ds);
 }
 
 ExifData *
@@ -788,21 +868,19 @@ exif_data_ref (ExifData *data)
 void
 exif_data_unref (ExifData *data)
 {
-	if (!data)
-		return;
+	if (!data) return;
 
 	data->priv->ref_count--;
-	if (!data->priv->ref_count)
-		exif_data_free (data);
+	if (!data->priv->ref_count) exif_data_free (data);
 }
 
 void
 exif_data_free (ExifData *data)
 {
 	unsigned int i;
+	ExifMem *mem = (data && data->priv) ? data->priv->mem : NULL;
 
-	if (!data)
-		return;
+	if (!data) return;
 
 	for (i = 0; i < EXIF_IFD_COUNT; i++) {
 		if (data->ifd[i]) {
@@ -810,19 +888,26 @@ exif_data_free (ExifData *data)
 			data->ifd[i] = NULL;
 		}
 	}
+
 	if (data->data) {
-		free (data->data);
+		exif_mem_free (mem, data->data);
 		data->data = NULL;
 	}
+
 	if (data->priv) {
+		if (data->priv->log) {
+			exif_log_unref (data->priv->log);
+			data->priv->log = NULL;
+		}
 		if (data->priv->md) {
 			exif_mnote_data_unref (data->priv->md);
 			data->priv->md = NULL;
 		}
-		free (data->priv);
-		data->priv = NULL;
+		exif_mem_free (mem, data->priv);
+		exif_mem_free (mem, data);
 	}
-	free (data);
+
+	exif_mem_unref (mem);
 }
 
 void
@@ -883,75 +968,11 @@ static void
 entry_set_byte_order (ExifEntry *e, void *data)
 {
 	ByteOrderChangeData *d = data;
-	unsigned int i;
-	ExifShort s;
-	ExifLong l;
-	ExifSLong sl;
-	ExifRational r;
-	ExifSRational sr;
 
 	if (!e)
 		return;
 
-	switch (e->format) {
-	case EXIF_FORMAT_SHORT:
-		for (i = 0; i < e->components; i++) {
-			s = exif_get_short (e->data +
-				(i * exif_format_get_size (e->format)),
-				d->old);
-			exif_set_short (e->data +
-				(i * exif_format_get_size (e->format)),
-				d->new, s);
-		}
-		break;
-	case EXIF_FORMAT_LONG:
-		for (i = 0; i < e->components; i++) {
-			l = exif_get_long (e->data +
-				(i * exif_format_get_size (e->format)),
-				d->old);
-			exif_set_long (e->data +
-				(i * exif_format_get_size (e->format)),
-				d->new, l);
-		}
-		break;
-	case EXIF_FORMAT_RATIONAL:
-		for (i = 0; i < e->components; i++) {
-			r = exif_get_rational (e->data +
-				(i * exif_format_get_size (e->format)),
-				d->old);
-			exif_set_rational (e->data +
-				(i * exif_format_get_size (e->format)),
-				d->new, r);
-		}
-		break;
-	case EXIF_FORMAT_SLONG:
-		for (i = 0; i < e->components; i++) {
-			sl = exif_get_slong (e->data +
-				(i * exif_format_get_size (e->format)),
-				d->old);
-			exif_set_slong (e->data +
-				(i * exif_format_get_size (e->format)),
-				d->new, sl);
-		}
-		break;
-	case EXIF_FORMAT_SRATIONAL:
-		for (i = 0; i < e->components; i++) {
-			sr = exif_get_srational (e->data +
-				(i * exif_format_get_size (e->format)),
-				d->old);
-			exif_set_srational (e->data +
-				(i * exif_format_get_size (e->format)),
-				d->new, sr);
-		}
-		break;
-	case EXIF_FORMAT_UNDEFINED:
-	case EXIF_FORMAT_BYTE:
-	case EXIF_FORMAT_SBYTE:
-	case EXIF_FORMAT_ASCII:
-	default:
-		/* Nothing here. */
-		break;
-	}
+	exif_array_set_byte_order (e->format, e->data, e->components, d->old, d->new);
 }
 
 static void
@@ -979,8 +1000,22 @@ exif_data_set_byte_order (ExifData *data, ExifByteOrder order)
 void
 exif_data_log (ExifData *data, ExifLog *log)
 {
+	unsigned int i;
+
 	if (!data || !data->priv) return;
 	exif_log_unref (data->priv->log);
 	data->priv->log = log;
 	exif_log_ref (log);
+
+	for (i = 0; i < EXIF_IFD_COUNT; i++)
+		exif_content_log (data->ifd[i], log);
+}
+
+/* Used internally within libexif */
+ExifLog *exif_data_get_log (ExifData *);
+ExifLog *
+exif_data_get_log (ExifData *data)
+{
+	if (!data || !data->priv) return NULL;
+	return data->priv->log;
 }
