@@ -18,49 +18,111 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include "config.h"
-#include "exif-entry.h"
-#include "exif-ifd.h"
-#include "exif-utils.h"
-#include "i18n.h"
+#include <config.h>
 
+#include <libexif/exif-entry.h>
+#include <libexif/exif-ifd.h>
+#include <libexif/exif-utils.h>
+#include <libexif/i18n.h>
+
+#include <ctype.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <time.h>
 #include <math.h>
 
-/* #define DEBUG */
-
-#undef  MIN
-#define MIN(a, b)  (((a) < (b)) ? (a) : (b))
-
 struct _ExifEntryPrivate
 {
 	unsigned int ref_count;
+
+	ExifMem *mem;
 };
 
+/* This function is hidden in exif-data.c */
+ExifLog *exif_data_get_log (ExifData *);
+
+static void
+exif_entry_log (ExifEntry *e, ExifLogCode code, const char *format, ...)
+{
+	va_list args;
+	ExifLog *l = NULL;
+
+	if (e && e->parent && e->parent->parent)
+		l = exif_data_get_log (e->parent->parent);
+	va_start (args, format);
+	exif_logv (l, code, "ExifEntry", format, args);
+	va_end (args);
+}
+
+static void *
+exif_entry_alloc (ExifEntry *e, unsigned int i)
+{
+	void *d;
+	ExifLog *l = NULL;
+
+	if (!e || !e->priv || !i) return NULL;
+
+	d = exif_mem_alloc (e->priv->mem, i);
+	if (d) return d;
+
+	if (e->parent && e->parent->parent)
+		l = exif_data_get_log (e->parent->parent);
+	EXIF_LOG_NO_MEMORY (l, "ExifEntry", i);
+	return NULL;
+}
+
+static void *
+exif_entry_realloc (ExifEntry *e, void *d_orig, unsigned int i)
+{
+	void *d;
+	ExifLog *l = NULL;
+
+	if (!e || !e->priv) return NULL;
+
+	if (!i) { exif_mem_free (e->priv->mem, d_orig); return NULL; }
+
+	d = exif_mem_realloc (e->priv->mem, d_orig, i);
+	if (d) return d;
+
+	if (e->parent && e->parent->parent)
+		l = exif_data_get_log (e->parent->parent);
+	EXIF_LOG_NO_MEMORY (l, "ExifEntry", i);
+	return NULL;
+}
+
 ExifEntry *
 exif_entry_new (void)
 {
-	ExifEntry *e;
+	ExifMem *mem = exif_mem_new_default ();
+	ExifEntry *e = exif_entry_new_mem (mem);
 
-	e = malloc (sizeof (ExifEntry));
+	exif_mem_unref (mem);
+
+	return e;
+}
+
+ExifEntry *
+exif_entry_new_mem (ExifMem *mem)
+{
+	ExifEntry *e = NULL;
+
+	e = exif_mem_alloc (mem, sizeof (ExifEntry));
 	if (!e) return NULL;
-	memset (e, 0, sizeof (ExifEntry));
-	e->priv = malloc (sizeof (ExifEntryPrivate));
-	if (!e->priv) { free (e); return NULL; }
-	memset (e->priv, 0, sizeof (ExifEntryPrivate));
+	e->priv = exif_mem_alloc (mem, sizeof (ExifEntryPrivate));
+	if (!e->priv) { exif_mem_free (mem, e); return NULL; }
 	e->priv->ref_count = 1;
 
-	return (e);
+	e->priv->mem = mem;
+	exif_mem_ref (mem);
+
+	return e;
 }
 
 void
 exif_entry_ref (ExifEntry *e)
 {
-	if (!e)
-		return;
+	if (!e) return;
 
 	e->priv->ref_count++;
 }
@@ -68,8 +130,7 @@ exif_entry_ref (ExifEntry *e)
 void
 exif_entry_unref (ExifEntry *e)
 {
-	if (!e)
-		return;
+	if (!e) return;
 
 	e->priv->ref_count--;
 	if (!e->priv->ref_count)
@@ -79,13 +140,198 @@ exif_entry_unref (ExifEntry *e)
 void
 exif_entry_free (ExifEntry *e)
 {
-	if (!e)
-		return;
+	if (!e) return;
 
-	if (e->data)
-		free (e->data);
-	free (e->priv);
-	free (e);
+	if (e->priv) {
+		ExifMem *mem = e->priv->mem;
+		if (e->data)
+			exif_mem_free (mem, e->data);
+		exif_mem_free (mem, e->priv);
+		exif_mem_free (mem, e);
+		exif_mem_unref (mem);
+	}
+}
+
+void
+exif_entry_fix (ExifEntry *e)
+{
+	unsigned int i;
+	ExifByteOrder o;
+	ExifRational r;
+	ExifSRational sr;
+
+	if (!e || !e->priv) return;
+
+	switch (e->tag) {
+	
+	/* These tags all need to be of format SHORT. */
+	case EXIF_TAG_YCBCR_SUB_SAMPLING:
+	case EXIF_TAG_SUBJECT_AREA:
+	case EXIF_TAG_COLOR_SPACE:
+	case EXIF_TAG_PLANAR_CONFIGURATION:
+	case EXIF_TAG_SENSING_METHOD:
+	case EXIF_TAG_ORIENTATION:
+	case EXIF_TAG_YCBCR_POSITIONING:
+	case EXIF_TAG_PHOTOMETRIC_INTERPRETATION:
+	case EXIF_TAG_CUSTOM_RENDERED:
+	case EXIF_TAG_EXPOSURE_MODE:
+	case EXIF_TAG_WHITE_BALANCE:
+	case EXIF_TAG_SCENE_CAPTURE_TYPE:
+	case EXIF_TAG_GAIN_CONTROL:
+	case EXIF_TAG_SATURATION:
+	case EXIF_TAG_CONTRAST:
+	case EXIF_TAG_SHARPNESS:
+		switch (e->format) {
+		case EXIF_FORMAT_LONG:
+			if (!e->parent || !e->parent->parent) break;
+			o = exif_data_get_byte_order (e->parent->parent);
+			for (i = 0; i < e->components; i++)
+				exif_set_short (
+					e->data + i *
+					exif_format_get_size (
+					EXIF_FORMAT_SHORT), o,
+					(ExifShort) exif_get_long (
+					e->data + i *
+					exif_format_get_size (
+					EXIF_FORMAT_LONG), o));
+			e->format = EXIF_FORMAT_SHORT;
+			e->size = e->components *
+				exif_format_get_size (e->format);
+			e->data = exif_entry_realloc (e, e->data, e->size);
+			exif_entry_log (e, EXIF_LOG_CODE_DEBUG,
+				"Tag '%s' was of format '%s' (which is "
+				"against specification) and has been "
+				"changed to format '%s'.",
+				exif_tag_get_name (e->tag), 
+				exif_format_get_name (EXIF_FORMAT_LONG),
+				exif_format_get_name (EXIF_FORMAT_SHORT));
+			break;
+		case EXIF_FORMAT_SHORT:
+		default:
+			break;
+		}
+		break;
+
+	/* All these tags need to be of format 'Rational'. */
+	case EXIF_TAG_FNUMBER:
+	case EXIF_TAG_APERTURE_VALUE:
+	case EXIF_TAG_EXPOSURE_TIME:
+	case EXIF_TAG_FOCAL_LENGTH:
+		switch (e->format) {
+		case EXIF_FORMAT_SRATIONAL:
+			if (!e->parent || !e->parent->parent) break;
+			o = exif_data_get_byte_order (e->parent->parent);
+			for (i = 0; i < e->components; i++) {
+				sr = exif_get_srational (e->data + i * 
+					exif_format_get_size (
+						EXIF_FORMAT_SRATIONAL), o);
+				r.numerator = (ExifLong) sr.numerator;
+				r.denominator = (ExifLong) sr.denominator;
+				exif_set_rational (e->data + i *
+					exif_format_get_size (
+						EXIF_FORMAT_RATIONAL), o, r);
+			}
+			e->format = EXIF_FORMAT_RATIONAL;
+			exif_entry_log (e, EXIF_LOG_CODE_DEBUG,
+				"Tag '%s' was of format '%s' (which is "
+				"against specification) and has been "
+				"changed to format '%s'.",
+				exif_tag_get_name (e->tag),
+				exif_format_get_name (EXIF_FORMAT_SRATIONAL),
+				exif_format_get_name (EXIF_FORMAT_RATIONAL));
+			break;
+		default:
+			break;
+		}
+		break;
+
+	case EXIF_TAG_USER_COMMENT:
+
+		/* Format needs to be UNDEFINED. */
+		if (e->format != EXIF_FORMAT_UNDEFINED) {
+			exif_entry_log (e, EXIF_LOG_CODE_DEBUG,
+				"Tag 'UserComment' had invalid format '%s'. "
+				"Format has been set to 'undefined'.",
+				exif_format_get_name (e->format));
+			e->format = EXIF_FORMAT_UNDEFINED;
+		}
+		/* Some packages like Canon ZoomBrowser EX 4.5 store
+		   only one zero byte followed by 7 bytes of rubbish */
+		if ((e->size >= 8) && (e->data[0] == 0)) {
+			memcpy(e->data, "\0\0\0\0\0\0\0\0", 8);
+		}
+
+		/* Some cameras fill the tag with '\0' or ' '. */
+		for (i = 0; i < e->size &&
+			    (!e->data[i] || (e->data[i] == ' ')); i++);
+		if (i && (i == e->size)) {
+			exif_entry_log (e, EXIF_LOG_CODE_DEBUG,
+				"Tag 'UserComment' contained unnecessary "
+				"data which has been removed.");
+			exif_mem_free (e->priv->mem, e->data);
+			e->data = NULL;
+			e->size = 0;
+			e->components = 0;
+		}
+
+		/* There need to be at least 8 bytes. */
+		if (e->size < 8) {
+			e->data = exif_entry_realloc (e, e->data, 8 + e->size);
+			if (!e->data) {
+				e->size = 0;
+				e->components = 0;
+				return;
+			}
+
+			/* Assume ASCII */
+			memmove (e->data + 8, e->data, e->size);
+			memcpy (e->data, "ASCII\0\0\0", 8);
+			e->size += 8;
+			exif_entry_log (e, EXIF_LOG_CODE_DEBUG,
+				"Tag 'UserComment' has been expanded to at "
+				"least 8 bytes in order to follow the "
+				"specification.");
+			break;
+		}
+
+		/*
+		 * If the first 8 bytes are empty and real data starts
+		 * afterwards, let's assume ASCII and claim the 8 first
+		 * bytes for the format specifyer.
+		 */
+		if (i >= 8) {
+			exif_entry_log (e, EXIF_LOG_CODE_DEBUG,
+				"Tag 'UserComment' did not start with "
+				"format identifyer. This has been fixed.");
+			memcpy (e->data, "ASCII\0\0\0", 8);
+		}
+
+		/* First 8 bytes need to follow the specification. */
+		if (memcmp (e->data, "ASCII\0\0\0"     , 8) &&
+		    memcmp (e->data, "UNICODE\0"       , 8) &&
+		    memcmp (e->data, "JIS\0\0\0\0\0"   , 8) &&
+		    memcmp (e->data, "\0\0\0\0\0\0\0\0", 8)) {
+			e->data = exif_entry_realloc (e, e->data, 8 + e->size);
+			if (!e->data) {
+				e->size = 0;
+				e->components = 0;
+				break;
+			}
+
+			/* Assume ASCII */
+			memmove (e->data + 8, e->data, e->size);
+			memcpy (e->data, "ASCII\0\0\0", 8);
+			e->size += 8;
+			exif_entry_log (e, EXIF_LOG_CODE_DEBUG,
+				"Tag 'UserComment' did not start with "
+				"format identifyer. This has been fixed.");
+			break;
+		}
+
+		break;
+	default:
+		break;
+	}
 }
 
 void
@@ -111,24 +357,27 @@ exif_entry_dump (ExifEntry *e, unsigned int indent)
 	printf ("%s  Value: %s\n", buf, exif_entry_get_value (e, value, sizeof(value)));
 }
 
-#define CF(format,target,v,maxlen)				\
-{								\
-	if (format != target) {					\
-		snprintf (v, maxlen,				\
-			_("Invalid format '%s', "		\
-			"expected '%s'."),			\
-			exif_format_get_name (format),		\
-			exif_format_get_name (target));		\
-		break;						\
-	}							\
+#define CF(entry,target,v,maxlen)					\
+{									\
+	if (entry->format != target) {					\
+		exif_entry_log (entry, EXIF_LOG_CODE_CORRUPT_DATA,	\
+			_("The tag '%s' contains data of an invalid "	\
+			"format ('%s', expected '%s')."),		\
+			exif_tag_get_name (entry->tag),			\
+			exif_format_get_name (entry->format),		\
+			exif_format_get_name (target));			\
+		break;							\
+	}								\
 }
 
-#define CC(number,target,v,maxlen)					\
+#define CC(entry,target,v,maxlen)					\
 {									\
-	if (number != target) {						\
-		snprintf (v, maxlen,					\
-			_("Invalid number of components (%i, "		\
-			"expected %i)."), (int) number, (int) target);	\
+	if (entry->components != target) {				\
+		exif_entry_log (entry, EXIF_LOG_CODE_CORRUPT_DATA,	\
+			_("The tag '%s' contains an invalid number of "	\
+			  "components (%i, expected %i)."),		\
+			exif_tag_get_name (entry->tag),		\
+			(int) entry->components, (int) target);		\
 		break;							\
 	}								\
 }
@@ -189,6 +438,7 @@ static struct {
       {  0, {NULL}}}},
   { EXIF_TAG_COMPRESSION,
     { {1, {N_("Uncompressed"), NULL}},
+      {5, {N_("LZW compression"), NULL}},
       {6, {N_("JPEG compression"), NULL}},
       {0, {NULL}}}},
   { EXIF_TAG_LIGHT_SOURCE,
@@ -336,18 +586,61 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 	case EXIF_TAG_USER_COMMENT:
 
 		/*
-		 * According to Ralf Holzer <rholzer@cmu.edu>,
+		 * The specification says UNDEFINED, but some
+		 * manufacturers don't care and use ASCII. If this is the
+		 * case here, only refuse to read it if there is no chance
+		 * of finding readable data.
+		 */
+		if ((e->format != EXIF_FORMAT_ASCII) || 
+		    (e->size <= 8) ||
+		    ( memcmp (e->data, "ASCII\0\0\0"  , 8) &&
+		      memcmp (e->data, "UNICODE\0"    , 8) &&
+		      memcmp (e->data, "JIS\0\0\0\0\0", 8) &&
+		      memcmp (e->data, "\0\0\0\0\0\0\0\0", 8)))
+			CF (e, EXIF_FORMAT_UNDEFINED, val, maxlen);
+
+		/*
+		 * Note that, according to the specification (V2.1, p 40),
 		 * the user comment field does not have to be 
 		 * NULL terminated.
 		 */
-		CF (e->format, EXIF_FORMAT_UNDEFINED, val, maxlen);
-		if (e->size < 8) break;
-		strncpy (val, e->data + 8, MIN (e->size - 8, maxlen));
+		if ((e->size >= 8) && !memcmp (e->data, "ASCII\0\0\0", 8)) {
+			strncpy (val, e->data + 8, MIN (e->size - 8, maxlen));
+			break;
+		}
+		if ((e->size >= 8) && !memcmp (e->data, "UNICODE\0", 8)) {
+			strncpy (val, _("Unsupported UNICODE string"), maxlen);
+			break;
+		}
+		if ((e->size >= 8) && !memcmp (e->data, "JIS\0\0\0\0\0", 8)) {
+			strncpy (val, _("Unsupported JIS string"), maxlen);
+			break;
+		}
+
+		/* Check if there is really some information in the tag. */
+		for (i = 0; (i < e->size) &&
+			    (!e->data[i] || (e->data[i] == ' ')); i++);
+		if (i == e->size) break;
+
+		/*
+		 * If we reach this point, the tag does not
+		 * comply with the standard and seems to contain data.
+		 * Print as much as possible.
+		 */
+		exif_entry_log (e, EXIF_LOG_CODE_DEBUG,
+			"Tag UserComment does not comply "
+			"with standard but contains data.");
+		for (; (i < e->size)  && (strlen (val) < maxlen - 1); i++) {
+			exif_entry_log (e, EXIF_LOG_CODE_DEBUG,
+				"Byte at position %i: 0x%02x", i, e->data[i]);
+			val[strlen (val)] =
+				isprint (e->data[i]) ? e->data[i] : '.';
+		}
 		break;
 
 	case EXIF_TAG_EXIF_VERSION:
-		CF (e->format, EXIF_FORMAT_UNDEFINED, val, maxlen);
-		CC (e->components, 4, val, maxlen);
+		CF (e, EXIF_FORMAT_UNDEFINED, val, maxlen);
+		CC (e, 4, val, maxlen);
 		strncpy (val, _("Unknown Exif Version"), maxlen);
 		for (i = 0; versions[i].label; i++) {
 			if (!memcmp (e->data, versions[i].label, 4)) {
@@ -360,8 +653,8 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 		}
 		break;
 	case EXIF_TAG_FLASH_PIX_VERSION:
-		CF (e->format, EXIF_FORMAT_UNDEFINED, val, maxlen);
-		CC (e->components, 4, val, maxlen);
+		CF (e, EXIF_FORMAT_UNDEFINED, val, maxlen);
+		CC (e, 4, val, maxlen);
 		if (!memcmp (e->data, "0100", 4))
 			strncpy (val, _("FlashPix Version 1.0"), maxlen);
 		else if (!memcmp (e->data, "0101", 4))
@@ -370,7 +663,7 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 			strncpy (val, _("Unknown FlashPix Version"), maxlen);
 		break;
 	case EXIF_TAG_COPYRIGHT:
-		CF (e->format, EXIF_FORMAT_ASCII, val, maxlen);
+		CF (e, EXIF_FORMAT_ASCII, val, maxlen);
 
 		/*
 		 * First part: Photographer.
@@ -390,7 +683,8 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 		    (strlen ((char *) e->data) + 1 < e->size) &&
 		    (strspn (e->data, " ") != strlen ((char *) e->data)))
 			strncat (val, e->data + strlen (e->data) + 1,
-				 maxlen - strlen (val));
+				 MIN (maxlen - strlen (val),
+				      e->size - (strlen (e->data) + 1)));
 		else
 			strncat (val, _("[None]"), maxlen - strlen (val));
 		strncat (val, " ", maxlen - strlen (val));
@@ -398,16 +692,16 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 
 		break;
 	case EXIF_TAG_FNUMBER:
-		CF (e->format, EXIF_FORMAT_RATIONAL, val, maxlen);
-		CC (e->components, 1, val, maxlen);
+		CF (e, EXIF_FORMAT_RATIONAL, val, maxlen);
+		CC (e, 1, val, maxlen);
 		v_rat = exif_get_rational (e->data, o);
 		if (!v_rat.denominator) return NULL;
 		snprintf (val, maxlen, "f/%.01f", (float) v_rat.numerator /
 						    (float) v_rat.denominator);
 		break;
 	case EXIF_TAG_APERTURE_VALUE:
-		CF (e->format, EXIF_FORMAT_RATIONAL, val, maxlen);
-		CC (e->components, 1, val, maxlen);
+		CF (e, EXIF_FORMAT_RATIONAL, val, maxlen);
+		CC (e, 1, val, maxlen);
 		v_rat = exif_get_rational (e->data, o);
 		if (!v_rat.denominator) return NULL;
 		snprintf (val, maxlen, "f/%.01f",
@@ -415,8 +709,8 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 				    (float) v_rat.denominator) / 2.));
 		break;
 	case EXIF_TAG_FOCAL_LENGTH:
-		CF (e->format, EXIF_FORMAT_RATIONAL, val, maxlen);
-		CC (e->components, 1, val, maxlen);
+		CF (e, EXIF_FORMAT_RATIONAL, val, maxlen);
+		CC (e, 1, val, maxlen);
 		v_rat = exif_get_rational (e->data, o);
 		if (!v_rat.denominator) return NULL;
 
@@ -451,16 +745,16 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 			strncat (val, b, maxlen - strlen (val) - 1);
 		break;
 	case EXIF_TAG_SUBJECT_DISTANCE:
-		CF (e->format, EXIF_FORMAT_RATIONAL, val, maxlen);
-		CC (e->components, 1, val, maxlen);
+		CF (e, EXIF_FORMAT_RATIONAL, val, maxlen);
+		CC (e, 1, val, maxlen);
 		v_rat = exif_get_rational (e->data, o);
 		if (!v_rat.denominator) return NULL;
 		snprintf (val, maxlen, "%.1f m", (float) v_rat.numerator /
 						   (float) v_rat.denominator);
 		break;
 	case EXIF_TAG_EXPOSURE_TIME:
-		CF (e->format, EXIF_FORMAT_RATIONAL, val, maxlen);
-		CC (e->components, 1, val, maxlen);
+		CF (e, EXIF_FORMAT_RATIONAL, val, maxlen);
+		CC (e, 1, val, maxlen);
 		v_rat = exif_get_rational (e->data, o);
 		if (!v_rat.denominator) return NULL;
 		d = (double) v_rat.numerator / (double) v_rat.denominator;
@@ -473,8 +767,8 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 			strncat (val, " sec.", maxlen - strlen (val) - 1);
 		break;
 	case EXIF_TAG_SHUTTER_SPEED_VALUE:
-		CF (e->format, EXIF_FORMAT_SRATIONAL, val, maxlen);
-		CC (e->components, 1, val, maxlen);
+		CF (e, EXIF_FORMAT_SRATIONAL, val, maxlen);
+		CC (e, 1, val, maxlen);
 		v_srat = exif_get_srational (e->data, o);
 		if (!v_srat.denominator) return NULL;
 		snprintf (val, maxlen, "%.0f/%.0f", (float) v_srat.numerator,
@@ -488,24 +782,24 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 			strncat (val, b, maxlen - strlen (val) - 1);
 		break;
 	case EXIF_TAG_BRIGHTNESS_VALUE:
-		CF (e->format, EXIF_FORMAT_SRATIONAL, val, maxlen);
-		CC (e->components, 1, val, maxlen);
+		CF (e, EXIF_FORMAT_SRATIONAL, val, maxlen);
+		CC (e, 1, val, maxlen);
 		v_srat = exif_get_srational (e->data, o);
 		snprintf (val, maxlen, "%i/%i", (int) v_srat.numerator,
 						  (int) v_srat.denominator);
 		/* FIXME: How do I calculate the APEX value? */
 		break;
 	case EXIF_TAG_FILE_SOURCE:
-		CF (e->format, EXIF_FORMAT_UNDEFINED, val, maxlen);
-		CC (e->components, 1, val, maxlen);
+		CF (e, EXIF_FORMAT_UNDEFINED, val, maxlen);
+		CC (e, 1, val, maxlen);
 		switch (e->data[0]) {
 		case 0x03: strncpy (val, _("DSC"), maxlen); break;
 		default: snprintf (val, maxlen, "0x%02x", e->data[0]); break;
 		}
 		break;
 	case EXIF_TAG_COMPONENTS_CONFIGURATION:
-		CF (e->format, EXIF_FORMAT_UNDEFINED, val, maxlen);
-		CC (e->components, 4, val, maxlen);
+		CF (e, EXIF_FORMAT_UNDEFINED, val, maxlen);
+		CC (e, 4, val, maxlen);
 		for (i = 0; i < 4; i++) {
 			switch (e->data[i]) {
 			case 0: c = _("-"); break;
@@ -522,8 +816,8 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 		}
 		break;
 	case EXIF_TAG_EXPOSURE_BIAS_VALUE:
-		CF (e->format, EXIF_FORMAT_SRATIONAL, val, maxlen);
-		CC (e->components, 1, val, maxlen);
+		CF (e, EXIF_FORMAT_SRATIONAL, val, maxlen);
+		CC (e, 1, val, maxlen);
 		v_srat = exif_get_srational (e->data, o);
 		if (!v_srat.denominator)
 			return NULL;
@@ -533,8 +827,8 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 			  (double) v_srat.denominator);
 		break;
 	case EXIF_TAG_YCBCR_SUB_SAMPLING:
-		CF (e->format, EXIF_FORMAT_SHORT, val, maxlen);
-		CC (e->components, 2, val, maxlen);
+		CF (e, EXIF_FORMAT_SHORT, val, maxlen);
+		CC (e, 2, val, maxlen);
 		v_short  = exif_get_short (e->data, o);
 		v_short2 = exif_get_short (
 			e->data + exif_format_get_size (e->format),
@@ -547,12 +841,12 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 			snprintf (val, maxlen, "%i, %i", v_short, v_short2);
 		break;
 	case EXIF_TAG_MAKER_NOTE:
-		CF (e->format, EXIF_FORMAT_UNDEFINED, val, maxlen);
+		CF (e, EXIF_FORMAT_UNDEFINED, val, maxlen);
 		snprintf (val, maxlen, _("%i bytes unknown data"),
 			  (int) e->components);
 		break;
 	case EXIF_TAG_SUBJECT_AREA:
-		CF (e->format, EXIF_FORMAT_SHORT, val, maxlen);
+		CF (e, EXIF_FORMAT_SHORT, val, maxlen);
 		switch (e->components) {
 		case 2:
 			v_short  = exif_get_short (e->data, o);
@@ -593,8 +887,8 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 	case EXIF_TAG_FLASH:
 	case EXIF_TAG_SUBJECT_DISTANCE_RANGE:
 	case EXIF_TAG_COLOR_SPACE:
-		CF (e->format,EXIF_FORMAT_SHORT, val, maxlen);
-		CC (e->components, 1, val, maxlen);
+		CF (e,EXIF_FORMAT_SHORT, val, maxlen);
+		CC (e, 1, val, maxlen);
 		v_short = exif_get_short (e->data, o);
 
 		/* Search the tag */
@@ -637,8 +931,8 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 	case EXIF_TAG_SATURATION:
 	case EXIF_TAG_CONTRAST:
 	case EXIF_TAG_SHARPNESS:
-		CF (e->format, EXIF_FORMAT_SHORT, val, maxlen);
-		CC (e->components, 1, val, maxlen);
+		CF (e, EXIF_FORMAT_SHORT, val, maxlen);
+		CC (e, 1, val, maxlen);
 		v_short = exif_get_short (e->data, o);
 
 		/* Search the tag */
@@ -690,13 +984,14 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 			break;
 		case EXIF_FORMAT_LONG:
 			v_long = exif_get_long (e->data, o);
-			snprintf (val, maxlen, "%i", (int) v_long);
+			snprintf (val, maxlen, "%li", (long int) v_long);
 			maxlen -= strlen (val);
 			for (i = 1; i < e->components; i++) {
 				v_long = exif_get_long (e->data +
 					exif_format_get_size (e->format) *
 					i, o);
-				snprintf (b, sizeof (b), ", %li", v_long);
+				snprintf (b, sizeof (b), ", %li",
+					(long int) v_long);
 				strncat (val, b, maxlen);
 				maxlen -= strlen (b);
 				if ((signed)maxlen <= 0) break;
@@ -704,13 +999,14 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 			break;
 		case EXIF_FORMAT_SLONG:
 			v_slong = exif_get_slong (e->data, o);
-			snprintf (val, maxlen, "%li", v_slong);
+			snprintf (val, maxlen, "%li", (long int) v_slong);
 			maxlen -= strlen (val);
 			for (i = 1; i < e->components; i++) {
 				v_long = exif_get_slong (e->data +
 					exif_format_get_size (e->format) *
 					i, o);
-				snprintf (b, sizeof (b), ", %li", v_long);
+				snprintf (b, sizeof (b), ", %li",
+						(long int) v_long);
 				strncat (val, b, maxlen);
 				maxlen -= strlen (b);
 				if ((signed)maxlen <= 0) break;
@@ -763,7 +1059,7 @@ exif_entry_get_value (ExifEntry *e, char *val, unsigned int maxlen)
 		}
 	}
 
-	return strlen (val) ? val : NULL;
+	return val;
 }
 
 void
@@ -793,8 +1089,7 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 1;
 		e->format = EXIF_FORMAT_LONG;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
-		if (e->data) memset (e->data, 0, e->size);
+		e->data = exif_entry_alloc (e, e->size);
 		break;
 
 	/* SHORT, 1 component, no default */
@@ -822,7 +1117,7 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 1;
 		e->format = EXIF_FORMAT_SHORT;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
+		e->data = exif_entry_alloc (e, e->size);
 		exif_set_short (e->data, o, 0);
 		break;
 
@@ -833,7 +1128,7 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 1;
 		e->format = EXIF_FORMAT_SHORT;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
+		e->data = exif_entry_alloc (e, e->size);
 		exif_set_short (e->data, o, 1);
 		break;
 
@@ -843,7 +1138,7 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 1;
 		e->format = EXIF_FORMAT_SHORT;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
+		e->data = exif_entry_alloc (e, e->size);
 		exif_set_short (e->data, o, 2);
 		break;
 
@@ -852,7 +1147,7 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 1;
 		e->format = EXIF_FORMAT_SHORT;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
+		e->data = exif_entry_alloc (e, e->size);
 		exif_set_short (e->data, o, 3);
 		break;
 
@@ -860,7 +1155,8 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 3;
 		e->format = EXIF_FORMAT_SHORT;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
+		e->data = exif_entry_alloc (e, e->size);
+		if (!e->data) break;
 		exif_set_short (e->data, o, 8);
 		exif_set_short (
 			e->data + exif_format_get_size (e->format),
@@ -873,7 +1169,8 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 2;
 		e->format = EXIF_FORMAT_SHORT;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
+		e->data = exif_entry_alloc (e, e->size);
+		if (!e->data) break;
 		exif_set_short (e->data, o, 2);
 		exif_set_short (
 			e->data + exif_format_get_size (e->format),
@@ -887,8 +1184,7 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 1;
 		e->format = EXIF_FORMAT_SRATIONAL;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
-		if (e->data) memset (e->data, 0, e->size);
+		e->data = exif_entry_alloc (e, e->size);
 		break;
 
 	/* RATIONAL, 1 component, no default */
@@ -908,8 +1204,7 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 1;
 		e->format = EXIF_FORMAT_RATIONAL;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
-		if (e->data) memset (e->data, 0, e->size);
+		e->data = exif_entry_alloc (e, e->size);
 		break;
 
 	/* RATIONAL, 1 component, default 72/1 */
@@ -918,7 +1213,7 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 1;
 		e->format = EXIF_FORMAT_RATIONAL;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
+		e->data = exif_entry_alloc (e, e->size);
 		r.numerator = 72;
 		r.denominator = 1;
 		exif_set_rational (e->data, o, r);
@@ -929,8 +1224,7 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 2;
 		e->format = EXIF_FORMAT_RATIONAL;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
-		if (e->data) memset (e->data, 0, e->size);
+		e->data = exif_entry_alloc (e, e->size);
 		break;
 
 	/* RATIONAL, 6 components */
@@ -938,7 +1232,8 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 6;
 		e->format = EXIF_FORMAT_RATIONAL;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
+		e->data = exif_entry_alloc (e, e->size);
+		if (!e->data) break;
 		r.denominator = 1;
 		r.numerator = 0;
 		exif_set_rational (e->data, o, r);
@@ -968,10 +1263,11 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 20;
 		e->format = EXIF_FORMAT_ASCII;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
+		e->data = exif_entry_alloc (e, e->size);
+		if (!e->data) break;
 		snprintf ((char *) e->data, e->size,
 			  "%04i:%02i:%02i %02i:%02i:%02i",
-			  tm->tm_year + 1900, tm->tm_mon, tm->tm_mday,
+			  tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
 			  tm->tm_hour, tm->tm_min, tm->tm_sec);
 		break;
 
@@ -992,14 +1288,16 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = strlen ("[None]") + 1;
 		e->format = EXIF_FORMAT_ASCII;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
+		e->data = exif_entry_alloc (e, e->size);
+		if (!e->data) break;
 		strncpy (e->data, "[None]", e->size);
 		break;
 	case EXIF_TAG_COPYRIGHT:
 		e->components = (strlen ("[None]") + 1) * 2;
 		e->format = EXIF_FORMAT_ASCII;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
+		e->data = exif_entry_alloc (e, e->size);
+		if (!e->data) break;
 		strcpy (e->data +                     0, "[None]");
 		strcpy (e->data + strlen ("[None]") + 1, "[None]");
 		break;
@@ -1010,7 +1308,7 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 0;
 		e->format = EXIF_FORMAT_UNDEFINED;
 		e->size = 0;
-		e->data = 0;
+		e->data = NULL;
 		break;
 
 	/* UNDEFINED, 1 component, default 1 */
@@ -1018,7 +1316,8 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 1;
 		e->format = EXIF_FORMAT_UNDEFINED;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
+		e->data = exif_entry_alloc (e, e->size);
+		if (!e->data) break;
 		e->data[0] = 0x01;
 		break;
 
@@ -1027,7 +1326,8 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
 		e->components = 1;
 		e->format = EXIF_FORMAT_UNDEFINED;
 		e->size = exif_format_get_size (e->format) * e->components;
-		e->data = malloc (e->size);
+		e->data = exif_entry_alloc (e, e->size);
+		if (!e->data) break;
 		e->data[0] = 0x03;
 		break;
 
@@ -1036,7 +1336,8 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
                 e->components = 4;
                 e->format = EXIF_FORMAT_UNDEFINED;
                 e->size = exif_format_get_size (e->format) * e->components;
-                e->data = malloc (e->size);
+                e->data = exif_entry_alloc (e, e->size);
+		if (!e->data) break;
                 memcpy (e->data, "0100", 4);
                 break;
 
@@ -1045,7 +1346,8 @@ exif_entry_initialize (ExifEntry *e, ExifTag tag)
                 e->components = 4;
                 e->format = EXIF_FORMAT_UNDEFINED;
                 e->size = exif_format_get_size (e->format) * e->components;
-                e->data = malloc (e->size);
+                e->data = exif_entry_alloc (e, e->size);
+		if (!e->data) break;
                 memcpy (e->data, "0210", 4);
                 break;
 
