@@ -18,18 +18,31 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include "config.h"
-#include "exif-tag.h"
-#include "i18n.h"
+#include <config.h>
+
+#include <libexif/exif-tag.h>
+#include <libexif/i18n.h>
 
 #include <stdlib.h>
+#include <string.h>
+
+typedef enum {
+	ESL_MANDATORY    = 1, /* Mandatory               */
+	ESL_CMANDATORY   = 2, /* Conditionally mandatory */
+	ESL_OPTIONAL     = 3, /* Optional                */
+	ESL_NOT_RECORDED = 4  /* Not recorded            */
+} ExifSL;                     /* Exif Support Level      */
 
 static struct {
 	ExifTag tag;
 	const char *name;
 	const char *title;
 	const char *description;
+	ExifSL esl_0[4], esl_1[4], esl_exif[4], esl_gps[4];
 } ExifTagTable[] = {
+	{EXIF_TAG_NEW_SUBFILE_TYPE, "NewSubfileType",
+	 "New Subfile Type", N_("A general indication of the kind of data "
+	    "contained in this subfile.")},
 	{EXIF_TAG_INTEROPERABILITY_INDEX, "InteroperabilityIndex",
 	 "InteroperabilityIndex",
 	 N_("Indicates the identification of the Interoperability rule. "
@@ -42,24 +55,29 @@ static struct {
 	{EXIF_TAG_IMAGE_WIDTH, "ImageWidth", N_("Image Width"),
 	 N_("The number of columns of image data, equal to the number of "
 	    "pixels per row. In JPEG compressed data a JPEG marker is "
-	    "used instead of this tag.")},
+	    "used instead of this tag."),
+	 {ESL_MANDATORY, ESL_MANDATORY, ESL_MANDATORY, ESL_NOT_RECORDED}},
 	{EXIF_TAG_IMAGE_LENGTH, "ImageLength", N_("Image Length"),
 	 N_("The number of rows of image data. In JPEG compressed data a "
-	    "JPEG marker is used instead of this tag.")},
+	    "JPEG marker is used instead of this tag."), 
+	 {ESL_MANDATORY, ESL_MANDATORY, ESL_MANDATORY, ESL_NOT_RECORDED}},
 	{EXIF_TAG_BITS_PER_SAMPLE, "BitsPerSample", N_("Bits per Sample"),
 	 N_("The number of bits per image component. In this standard each "
 	    "component of the image is 8 bits, so the value for this "
 	    "tag is 9. See also <SamplesPerPixel>. In JPEG compressed data "
-	    "a JPEG marker is used instead of this tag.")},
+	    "a JPEG marker is used instead of this tag."),
+	 {ESL_MANDATORY, ESL_MANDATORY, ESL_MANDATORY, ESL_NOT_RECORDED}},
 	{EXIF_TAG_COMPRESSION, "Compression", N_("Compression"),
 	 N_("The compression scheme used for the image data. When a "
 	    "primary image is JPEG compressed, this designation is "
 	    "not necessary and is omitted. When thumbnails use JPEG "
-	    "compression, this tag value is set to 6.")},
+	    "compression, this tag value is set to 6."),
+	 {ESL_MANDATORY, ESL_MANDATORY, ESL_MANDATORY, ESL_NOT_RECORDED}},
 	{EXIF_TAG_PHOTOMETRIC_INTERPRETATION, "PhotometricInterpretation",
 	 N_("Photometric Interpretation"),
 	 N_("The pixel composition. In JPEG compressed data a JPEG "
-	    "marker is used instead of this tag.")},
+	    "marker is used instead of this tag."),
+	 {ESL_MANDATORY, ESL_MANDATORY, ESL_MANDATORY, ESL_NOT_RECORDED}},
 	{EXIF_TAG_FILL_ORDER, "FillOrder", N_("Fill Order"), ""},
 	{EXIF_TAG_DOCUMENT_NAME, "DocumentName", N_("Document Name"), ""},
 	{EXIF_TAG_IMAGE_DESCRIPTION, "ImageDescription",
@@ -68,26 +86,31 @@ static struct {
 	    "a comment such as \"1988 company picnic\" or "
 	    "the like. Two-bytes character codes cannot be used. "
 	    "When a 2-bytes code is necessary, the Exif Private tag "
-	    "<UserComment> is to be used.")},
+	    "<UserComment> is to be used."),
+	 {ESL_CMANDATORY, ESL_CMANDATORY, ESL_CMANDATORY, ESL_CMANDATORY}},
 	{EXIF_TAG_MAKE, "Make", N_("Manufacturer"),
 	 N_("The manufacturer of the recording "
 	    "equipment. This is the manufacturer of the DSC, scanner, "
 	    "video digitizer or other equipment that generated the "
 	    "image. When the field is left blank, it is treated as "
-	    "unknown.")},
+	    "unknown."),
+	 {ESL_CMANDATORY, ESL_CMANDATORY, ESL_CMANDATORY, ESL_CMANDATORY}},
 	{EXIF_TAG_MODEL, "Model", N_("Model"),
 	 N_("The model name or model number of the equipment. This is the "
 	    "model name or number of the DSC, scanner, video digitizer "
 	    "or other equipment that generated the image. When the field "
-	    "is left blank, it is treated as unknown.")},
+	    "is left blank, it is treated as unknown."),
+	 {ESL_CMANDATORY, ESL_CMANDATORY, ESL_CMANDATORY, ESL_CMANDATORY}},
 	{EXIF_TAG_STRIP_OFFSETS, "StripOffsets", N_("Strip Offsets"),
 	 N_("For each strip, the byte offset of that strip. It is "
 	    "recommended that this be selected so the number of strip "
 	    "bytes does not exceed 64 Kbytes. With JPEG compressed "
 	    "data this designation is not needed and is omitted. See also "
-	    "<RowsPerStrip> and <StripByteCounts>.")},
+	    "<RowsPerStrip> and <StripByteCounts>."),
+	 {ESL_CMANDATORY, ESL_CMANDATORY, ESL_CMANDATORY, ESL_CMANDATORY}},
 	{EXIF_TAG_ORIENTATION, "Orientation", N_("Orientation"),
-	 N_("The image orientation viewed in terms of rows and columns.")},
+	 N_("The image orientation viewed in terms of rows and columns."),
+	 {ESL_CMANDATORY, ESL_CMANDATORY, ESL_CMANDATORY, ESL_CMANDATORY}},
 	{EXIF_TAG_SAMPLES_PER_PIXEL, "SamplesPerPixel",
 	 N_("Samples per Pixel"),
 	 N_("The number of components per pixel. Since this standard applies "
@@ -151,6 +174,8 @@ static struct {
 	    "Normally this tag is not necessary, since colorspace is "
 	    "specified in the colorspace information tag (<ColorSpace>).")},
 	{EXIF_TAG_TRANSFER_RANGE, "TransferRange", N_("Transfer Range"), ""},
+	{EXIF_TAG_SUB_IFDS, "SubIFDs", "SubIFD Offsets", N_("Defined by Adobe Corporation "
+	    "to enable TIFF Trees within a TIFF file.")},
 	{EXIF_TAG_JPEG_PROC, "JPEGProc", "JPEGProc", ""},
 	{EXIF_TAG_JPEG_INTERCHANGE_FORMAT, "JPEGInterchangeFormat",
 	 N_("JPEG Interchange Format"),
@@ -203,6 +228,7 @@ static struct {
 	    "in a color space information tag, with the default "
 	    "being the value that gives the optimal image characteristics "
 	    "Interoperability these conditions.")},
+	{EXIF_TAG_XML_PACKET, "XMLPacket", N_("XML Packet"), N_("XMP Metadata")},
 	{EXIF_TAG_RELATED_IMAGE_FILE_FORMAT, "RelatedImageFileFormat",
 	 "RelatedImageFileFormat", ""},
 	{EXIF_TAG_RELATED_IMAGE_WIDTH, "RelatedImageWidth",
@@ -243,6 +269,7 @@ static struct {
 	{EXIF_TAG_FNUMBER, "FNumber", "FNumber",
 	 N_("The F number.")},
 	{EXIF_TAG_IPTC_NAA, "IPTC/NAA", "IPTC/NAA", ""},
+	{EXIF_TAG_IMAGE_RESOURCES, "ImageResources", N_("Image Resources Block"), ""},
 	{EXIF_TAG_EXIF_IFD_POINTER, "ExifIFDPointer", "ExifIFDPointer",
 	 N_("A pointer to the Exif IFD. Interoperability, Exif IFD has the "
 	    "same structure as that of the IFD specified in TIFF. "
@@ -514,6 +541,7 @@ static struct {
 	{EXIF_TAG_SUBJECT_AREA, "SubjectArea", N_("Subject Area"),
 	 N_("This tag indicates the location and area of the main subject "
 	    "in the overall scene.")},
+	{EXIF_TAG_TIFF_EP_STANDARD_ID, "TIFF/EPStandardID", N_("TIFF/EP Standard ID"), ""},
 	{EXIF_TAG_CUSTOM_RENDERED, "CustomRendered", N_("Custom Rendered"),
 	 N_("This tag indicates the use of special processing on image "
 	    "data, such as rendering geared to output. When special "
@@ -603,7 +631,7 @@ exif_tag_get_name (ExifTag tag)
 		if (ExifTagTable[i].tag == tag)
 			break;
 
-	return (ExifTagTable[i].name);
+	return ExifTagTable[i].name;
 }
 
 const char *
@@ -615,10 +643,8 @@ exif_tag_get_title (ExifTag tag)
 	bindtextdomain (GETTEXT_PACKAGE, LIBEXIF_LOCALEDIR);
 
 	for (i = 0; ExifTagTable[i].title; i++)
-		if (ExifTagTable[i].tag == tag)
-			break;
-
-	return (_(ExifTagTable[i].title));
+		if (ExifTagTable[i].tag == tag) break;
+	return _(ExifTagTable[i].title);
 }
 
 const char *
@@ -630,8 +656,18 @@ exif_tag_get_description (ExifTag tag)
 	bindtextdomain (GETTEXT_PACKAGE, LIBEXIF_LOCALEDIR);
 
 	for (i = 0; ExifTagTable[i].description; i++)
-		if (ExifTagTable[i].tag == tag)
-			break;
+		if (ExifTagTable[i].tag == tag) break;
+	return _(ExifTagTable[i].description);
+}
 
-	return (_(ExifTagTable[i].description));
+ExifTag 
+exif_tag_from_name (const char *name)
+{
+	unsigned int i;
+
+	if (!name) return 0;
+
+	for (i = 0; ExifTagTable[i].name; i++)
+		if (!strcmp (ExifTagTable[i].name, name)) break;
+	return ExifTagTable[i].tag;
 }
